{****************************************************************************
** Serial communications unit                                              **
**  by Steven H Don                                                        **
*****************************************************************************
** Unit for communicating with another computer via modem or serial cable. **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************}
unit COMMUnit;

interface

{
  Functions available to the calling program are:
    COMOpenPort     : Opens a COM port for read/write access
    COMClosePort    : Closes the COM port that was opened with the above command.
    COMReadByte     : Reads a byte from the buffer, if available
                      Returns TRUE if successful.
    COMWriteByte    : Writes a byte to the COM port.
                      Returns TRUE if successful.
    COMWriteString  : Writes a string of bytes to the COM port.
                      Returns TRUE if successful.
    COMCarrier      : Returns TRUE if carrier detect.
}

procedure COMOpenPort (p, i : Byte; b : LongInt);
procedure COMClosePort;
function  COMReadByte (var b : Byte) : Boolean;
function  COMWriteByte (b : Byte) : Boolean;
function  COMWriteString (t : String) : Boolean;
function  COMCarrier : Boolean;

implementation

uses DOS;

type BufferType = Array [0..1023] of Byte;

{Local variables}
var
  {This stores the address of the old IRQ handler}
  OldVec             : Pointer;
  {This stores data on the port currently open}
  CAddr, CIRQ        : Word;
  {This is used to store the incoming data}
  Buffer             : ^BufferType;
  ReadPtr, WritePtr  : Word;
  {UART registers must be saved}
  UARTRegs           : Array [0..1] of Byte;

{
  Local functions and procedures:
    COMAddress : Returns the port address of a COM port.
    IRQHandler : Handles the reception of incoming data.
}

{This returns the hardware port address of the COM port, as reported
by the BIOS}
function COMAddress (p : Byte) : Word;
begin
  COMAddress := MemW [$0040:(p - 1) shl 1];
end;

{This handles incoming data and stores it in the buffer}
procedure IRQHandler; interrupt;
begin
  {Enable interrupts}
  asm sti end;
  {Store byte in buffer}
  Buffer^ [WritePtr] := Port [CAddr];
  {Move on to next byte in buffer}
  WritePtr := (WritePtr + 1) and 1023;
  {Return interrupts to disable}
  asm cli end;
  {Acknowledge hardware interrupt}
  Port [$20] := $20;
end;

{This opens a COM port for read/write access, it:
- initializes the buffer
- installs the new IRQ handler
- stores the UART registers and reprograms the UART}
procedure COMOpenPort (p, i : Byte; b : LongInt);
var
  BaudRateDivider : LongInt;

begin
  {Assign buffer}
  New (Buffer);
  {Reset read/write pointers}
  ReadPtr := 0; WritePtr := 0;

  {Retrieve address}
  CAddr := COMAddress (p);

  {Store UART registers}
  UARTRegs [0] := Port [CAddr + 1];
  UARTRegs [1] := Port [CAddr + 4];

  {Disable interrupts}
  asm cli end;
  CIRQ := i;
  {Get old handler}
  GetIntVec (CIRQ + 8, OldVec);
  {Save new handler}
  SetIntVec (CIRQ + 8, Addr (IRQHandler));
  {Set UART registers}
  Port [CAddr + 4] := 8;
  Port [CAddr + 1] := 1;
  {Enable IRQ}
  Port[$21] := Port[$21] and not (1 shl CIRQ);
  {Enable interrupts}
  asm sti end;

  {Initialize the UART}
  {set baudrate}
  BaudRateDivider := 115200 div b;
  Port [CAddr + 3] := $80;
  Port [CAddr    ] := lo (BaudRateDivider);
  Port [CAddr + 1] := hi (BaudRateDivider);
  {set to 1 stopbit, 8 databits, no parity}
  Port [CAddr + 3] := 3;

  {Relieve the modem}
  if Port [CAddr    ] = 0 then;
  {Reset line status register}
  if Port [CAddr + 5] = 0 then;
  {Enable DTR = Data Terminal Ready}
  Port [CAddr + 4] := Port [CAddr + 4] or 1;
end;

{This closes the COM port, undoing COMOpenPort}
procedure COMClosePort;
begin
  {Disable interrupts}
  asm cli end;
  {Mask IRQ}
  Port[$21] := Port[$21] or (1 shl CIRQ);
  {Restore UART registers}
  Port [CAddr + 1] := UARTRegs [0];
  Port [CAddr + 4] := UARTRegs [1];
  {Restore old handler}
  SetIntVec (CIRQ + 8, OldVec);
  {Enable interrupts}
  asm sti end;
end;

{This checks whether a byte is available.
If a byte was found, it returns true and "b" contains the value of that byte

If no byte was found, it returns false.
WARNING: The values returned by "b" in this case have no meaning}
function COMReadByte (var b : Byte) : Boolean;
begin
  {Assume no byte is waiting}
  COMReadByte := false;
  {If there is a byte waiting in the buffer, return true and store it}
  if ReadPtr <> WritePtr then begin
    COMReadByte := true;
    b := Buffer^ [ReadPtr];
    {Increase pointer}
    ReadPtr := (ReadPtr + 1) and 1023;
  end;
end;

{This writes a byte to the COM port. It returns true if successful, false if
the port didn't accept the write}
function COMWriteByte (b : Byte) : Boolean;
var
  TimeOut : Word;

begin
  {Wait until the last byte was sent}
  while Port [CAddr + 5] and $20 = 0 do;
  {Wait for COM port to indicate CTS = Clear To Send, stopping if the
  connection has been dropped}
  TimeOut := MaxInt;
  Port [CAddr + 4] := $B;
  while (Port [CAddr + 6] and $10 = 0) and (TimeOut > 0) do dec (TimeOut);
  {If there was no timeout, send the character, otherwise, report failure}
  if TimeOut = 0 then
    COMWriteByte := false
  else begin
    Port [CAddr] := b;
    COMWriteByte := true;
  end;
end;

{This sends an entire string of characters to the port using the above
procedure}
function COMWriteString (t : String) : Boolean;
var
  Position : Byte;

begin
  {Assume failure}
  COMWriteString := false;
  for Position := 1 to length (t) do begin
    if not COMWriteByte (Ord (t [Position])) then Exit;
  end;
  {Successful}
  COMWriteString := true;
end;

{This returns true if a carrier is detected by the COM port}
function COMCarrier : Boolean;
begin
  COMCarrier := Port [CAddr + 6] and $80 = $80;
end;

end.